#!/usr/bin/python
# -*- coding: utf-8 -*-

# attention: pour les accents, ce fichier doit être impérativement
#            lu et modifié sous éditeur avec encodage UTF-8

################################################################
######                                                    ######
######       Module additionnel: probabilite.py           ######
######                                                    ######
######            accompagnant la CALCULEXT               ######
######      [Calculatrice Extensible Version 1.40]        ######
######                                                    ######
################################################################

# Copyright (C) 2008 Jean-Paul Vidal dit "Tyrtamos" — Tous droits réservés.
#   Logiciel distribué sur le site <http://linux.jpvweb.com>
#
#   Ce programme est un logiciel libre. Vous pouvez le redistribuer ou le
#   modifier suivant les termes de la "GNU General Public License" version 3,
#   telle que publiée par la Free Software Foundation.
#
#   Ce programme est distribué dans l’espoir qu’il vous sera utile, mais SANS
#   AUCUNE GARANTIE : sans même la garantie implicite de COMMERCIALISABILITÉ
#   ni d’ADÉQUATION À UN OBJECTIF PARTICULIER. Consultez la Licence Générale
#   Publique GNU pour plus de détails.
#
#   Ce programme doit être distribué avec une copie de la
#   License Générale Publique GNU (fichier license.html).
#   Si ce n’est pas le cas, consultez: <http://www.gnu.org/licenses/gpl.html>.

##################################################################
# Attention: configuration volontairement inhabituelle pour python
# les divisions entre 2 entiers seront décimales et pas entières!
# par exemple: 10/3 donnera "3.33333333333333" et pas "3"
# (les divisions entières seront désormais faites avec "//")
#
from __future__ import division

##################################################################
# Importation des modules système
#
import time

from sys import modules
pmain = modules['__main__']

##################################################################
# Importation des modules nécessaires aux fonctions du présent module
#   (à compléter si nécessaire)
#
from math import *
from random import *
from combinatoire import *

###############################################################################
# Loi binomiale
#   probabilité d'avoir k réussite(s) dans n évènements indépendants,
#     chaque évènement ayant une probabilité p de réussite.
#
def binom(k,n,p):
    """binom(k,n,p): probabilité d'avoir k réussite(s) dans n évènements indépendants, chaque évènement ayant une probabilité p% de réussite"""
    # vérification des données
    if (not pmain.estnombre(k)) or (not pmain.estentier(k)) or (k<0):
        raise ValueError ("erreur binom(k,n,p): k doit être un nombre entier positif ou nul")
    if (not pmain.estnombre(n)) or (not pmain.estentier(n)) or (n<=0):
        raise ValueError ("erreur binom(k,n,p): n doit être un nombre entier positif")
    if (not pmain.estnombre(p)) or (p<0) or (p>1):
        raise ValueError ("erreur binom(k,n,p): p doit être un nombre compris entre 0 et 1")
    if (k>n):
        raise ValueError ("erreur binom(k,n,p): k doit être inférieur ou égal à n")
    # calcul
    if k==n:
        return p**n
    if p==0:
        return 0.0
    if p==1.0:
        # NB: le cas k==n a déjà été traité plus haut
        return 0.0
    q=1-p
    if k==0:
        return q**n
    j1=k
    j2=n-k
    xmin=1.0e-200
    x=1.0
    for i in xrange(0,k):
        x*=n/k
        while (x>xmin) and (j1>0):
            x*=p
            j1-=1
        while (x>xmin) and (j2>0):
            x*=q
            j2-=1
        n-=1
        k-=1
    return x*(p**j1)*(q**j2)


###############################################################################
# Loi binomiale cumulée
#   prob d'avoir 0,1,..,k réussite(s) dans n évènements indépendants,
#     chaque évènement ayant une probabilité p de réussite.
#
def binomcum(k,n,p):
    """binomcum(k,n,p): probabilité d'avoir 0,1,..,k réussite(s) dans n évènements indépendants, chaque évènement ayant une probabilité p% de réussite"""
    # vérification des données
    if (not pmain.estnombre(k)) or (not pmain.estentier(k)) or (k<0):
        raise ValueError ("erreur binomcum(k,n,p): k doit être un nombre entier positif ou nul")
    if (not pmain.estnombre(n)) or (not pmain.estentier(n)) or (n<=0):
        raise ValueError ("erreur binomcum(k,n,p): n doit être un nombre entier positif")
    if (not pmain.estnombre(p)) or (p<0) or (p>1):
        raise ValueError ("erreur binomcum(k,n,p): p doit être un nombre compris entre 0 et 1")
    if (k>n):
        raise ValueError ("erreur binomcum(k,n,p): k doit être inférieur ou égal à n")
    # calcul
    if k==n:
        return 1.0
    if p==0:
        return 0.0
    if p==1.0:
        # NB: le cas k==n a déjà été traité plus haut
        return 0.0
    j=int(p*n)
    b=0
    xmin=1.0e-20
    if k<=j:
        for i in xrange(k,-1,-1):
            x=binom(i,n,p)
            b+=x
            if x<xmin:
                break
    else:
        for i in xrange(k+1,n+1):
            x=binom(i,n,p)
            b+=x
            if x<xmin:
                break
        b=1.0-b
    return b

###############################################################################
# Loi binomiale: borne supérieure de confiance sur la proportion donnée par un échantillon (k,n) avec une confiance conf
#
def binomconfsup(k,n,conf):
    """binomconfsup(k,n,conf): borne supérieure de confiance sur la proportion donnée par un échantillon (k,n) avec une confiance conf"""
    # vérification des données
    if (not pmain.estnombre(k)) or (not pmain.estentier(k)) or (k<0):
        raise ValueError ("erreur binomconfsup(k,n,conf): k doit être un nombre entier positif ou nul")
    if (not pmain.estnombre(n)) or (not pmain.estentier(n)) or (n<=0):
        raise ValueError ("erreur binomconfsup(k,n,conf): n doit être un nombre entier positif")
    if (not pmain.estnombre(conf)) or (conf<0) or (conf>1):
        raise ValueError ("erreur binomconfsup(k,n,conf): conf doit être un nombre compris entre 0 et 1")
    if (k>n):
        raise ValueError ("erreur binomconfsup(k,n,conf): k doit être inférieur ou égal à n")
    # calcul
    eps=1.0e-8
    r=1.0-conf
    p1=k/n
    dp=0.1
    f1=binomcum(k,n,p1)-r
    while True:
        p2=p1+dp
        f2=binomcum(k,n,p2)-r
        if abs(f2)<eps:
            break
        if ((f1>0)and(f2>0))or((f1<0)and(f2<0)):
            if abs(f2)>abs(f1):
                dp*=-0.5
        else:
            dp*=-0.5
        p1=p2
        f1=f2
    return p2

###############################################################################
# Loi binomiale: borne inférieure de confiance sur la proportion donnée par un échantillon (k,n) avec une confiance conf
#
def binomconfinf(k,n,conf):
    """binomconfinf(k,n,conf): borne inférieure de confiance sur la proportion donnée par un échantillon (k,n) avec une confiance conf"""
    # vérification des données
    if (not pmain.estnombre(k)) or (not pmain.estentier(k)) or (k<0):
        raise ValueError ("erreur binomconfinf(k,n,conf): k doit être un nombre entier positif ou nul")
    if (not pmain.estnombre(n)) or (not pmain.estentier(n)) or (n<=0):
        raise ValueError ("erreur binomconfinf(k,n,conf): n doit être un nombre entier positif")
    if (not pmain.estnombre(conf)) or (conf<0) or (conf>1):
        raise ValueError ("erreur binomconfinf(k,n,conf): conf doit être un nombre compris entre 0 et 1")
    if (k>n):
        raise ValueError ("erreur binomconfinf(k,n,conf): k doit être inférieur ou égal à n")
    # calcul
    eps=1.0e-8
    r=1-conf
    dp=-0.1
    p1=k/n
    f1=1-binomcum(k-1,n,p1)-r
    while True:
        p2=p1+dp
        f2=1-binomcum(k-1,n,p2)-r
        if abs(f2)<eps:
            break
        if ((f1>0)and(f2>0))or((f1<0)and(f2<0)):
            if abs(f2)>abs(f1):
                dp*=-0.5
        else:
            dp*=-0.5
        p1=p2
        f1=f2
    return p2

###############################################################################
# Loi binomiale: intervalle de confiance sur la proportion donnée par un échantillon (k,n) avec une confiance conf
#
def binomconf(k,n,conf):
    """binomconf(k,n,conf): intervalle de confiance sur la proportion donnée par un échantillon (k,n) avec une confiance conf"""
    # vérification des données
    if (not pmain.estnombre(k)) or (not pmain.estentier(k)) or (k<0):
        raise ValueError ("erreur binomconf(k,n,conf): k doit être un entier positif ou nul")
    if (not pmain.estnombre(n)) or (not pmain.estentier(n)) or (n<=0):
        raise ValueError ("erreur binomconf(k,n,conf): n doit être un entier positif")
    if (not pmain.estnombre(conf)) or (conf<0) or (conf>1):
        raise ValueError ("erreur binomconf(k,n,conf): conf doit être un nombre compris entre 0 et 1")
    if (k>n):
        raise ValueError ("erreur binomconf(k,n,conf): k doit être <= n")
    # calcul
    bi=binomconfinf(k,n,conf+(1-conf)/2)
    bs=binomconfsup(k,n,conf+(1-conf)/2)
    return [bi,bs]

###############################################################################
# Loi binomiale: valeur(s) tirée(s) au hasard selon une distribution binomiale
#
def hbinom(n,p,nb=0):
    """hbinom(n,p,nb=0): valeur(s) tirée(s) au hasard selon une distribution binomiale"""
    def _hbinom(n,p):
        ph=random()
        i=0
        pc=binom(i,n,p)
        while ph>pc:
            i+=1
            pc+=binom(i,n,p)
        return i
    # vérification des données
    if (not pmain.estnombre(n)) or (not pmain.estentier(n)) or (n<=0):
        raise ValueError ("erreur hbinom(n,p,nb=0): n doit être un nombre entier positif")
    if (not pmain.estnombre(p)) or (p<0) or (p>1):
        raise ValueError ("erreur hbinom(n,p,nb=0): p doit être un nombre compris entre 0 et 1")
    if (not pmain.estnombre(nb)) or (not pmain.estentier(nb)) or (nb<0):
        raise ValueError ("erreur hbinom(n,p,nb=0): nb doit être un nombre entier positif")
    # calcul
    if nb==0:
        return _hbinom(n,p)
    else:
        R=[]
        for j in xrange(0,nb):
            R.append(_hbinom(n,p))
        return R

###############################################################################
# Loi hypergéométrique
#   probabilité d'avoir k réussite(s) dans n évènements indépendants,
#     sachant qu'il y a g de réussites dans la population de taille t.
#
def hypgeo(k,n,g,t):
    """hypgeo(k,n,g,t): prob d'avoir k réussites dans un échantillon de taille n, sachant qu'il y en a g la population de taille t"""
    # vérification des données
    if (not pmain.estnombre(k)) or (not pmain.estentier(k)) or (k<0):
        raise ValueError ("erreur hypgeo(k,n,p,t): k doit être un entier positif ou nul")
    if (not pmain.estnombre(n)) or (not pmain.estentier(n)) or (n<=0):
        raise ValueError ("erreur hypgeo(k,n,p,t): n doit être un entier positif")
    if (not pmain.estnombre(g)) or (not pmain.estentier(g)) or (g<0):
        raise ValueError ("erreur hypgeo(k,n,p,t): g doit être un entier positif ou nul")
    if (not pmain.estnombre(t)) or (not pmain.estentier(t)) or (t<0):
        raise ValueError ("erreur hypgeo(k,n,p,t): t doit être un entier positif")
    if (k>n):
        raise ValueError ("erreur hypgeo(k,n,g,t): k doit être <= n")
    if (n>t):
        raise ValueError ("erreur hypgeo(k,n,g,t): n doit être <= t")
    if (g>t):
        raise ValueError ("erreur hypgeo(k,n,g,t): g doit être <= t")
    if (k>g):
        raise ValueError ("erreur hypgeo(k,n,g,t): k doit être <= g")
    # calcul
    return combin(g,k)*combin(t-g,n-k)/combin(t,n)

###############################################################################
# Loi hypergéometrique cumulée
#   probabilité d'avoir 0,1,2,...,k réussite(s) dans n évènements indépendants,
#     sachant qu'il y a g de réussites dans la population de taille t.
#
def hypgeocum(k,n,g,t):
    """hypgeocum(k,n,p,t): prob d'avoir 0,1,..,k réussites dans un échantillon de taille n, sachant qu'il y p% de réussite dans la population de taille t"""
    # vérification des données
    if (not pmain.estnombre(k)) or (not pmain.estentier(k)) or (k<0):
        raise ValueError ("erreur hypgeocum(k,n,p,t): k doit être un entier positif ou nul")
    # calcul
    x=0
    for i in range(0,k+1):
        x+=hypgeo(i,n,g,t)
    return x

###############################################################################
# probabilité d'avoir k évènements distribués selon une loi de Poisson de paramètre m
#
def poisson(k,m):
    """poisson(k,m): probabilité d'avoir k évènements distribués selon une loi de Poisson de paramètre m"""
    if (not pmain.estnombre(k)) or (not pmain.estentier(k)) or (k<0):
        raise ValueError ("erreur poisson(k,m): k doit être un nombre entier positif ou nul")
    if (not pmain.estnombre(m)) or (m<0):
        raise ValueError ("erreur poisson(k,m): m doit être un nombre positif ou nul")
    p=e**(-m)
    for i in xrange(0,k):
        p*=m/k
        k-=1
    return p

###############################################################################
# probabilité d'avoir 0,1,..,k évènements distribués selon une loi de Poisson de paramètre m
#
def poissoncum(k,m):
    """poissoncum(k,m): probabilité d'avoir 0,1,..,k évènement(s) distribués selon une loi de Poisson de paramètre m"""
    if (not pmain.estnombre(k)) or (not pmain.estentier(k)) or (k<0):
        raise ValueError ("erreur poissoncum(k,m): k doit être un nombre entier positif ou nul")
    if (not pmain.estnombre(m)) or (m<0):
        raise ValueError ("erreur poissoncum(k,m): m doit être un nombre positif ou nul")
    pk=e**(-m)
    p=pk
    for i in xrange(1,k+1):
        pk*=m/i
        p+=pk
    return p

###############################################################################
# Borne supérieure de confiance sur le paramètre m d'une distribution de Poisson
#
def poissonconfsup(k,conf):
    """poissonconfsup(k,conf): Borne supérieure de confiance sur le paramètre m d'une distribution de Poisson"""
    if (not pmain.estnombre(k)) or (not pmain.estentier(k)) or (k<0):
        raise ValueError ("erreur poissonconfsup(k,conf): k doit être un nombre entier positif ou nul")
    if (not pmain.estnombre(conf)) or (conf<0) or (conf>1):
        raise ValueError ("erreur poissonconfsup(k,conf): conf doit être un nombre compris entre 0 et 1")
    dp=0.1
    eps=1.0e-8
    m1=0.5
    r=1-conf
    f1=poissoncum(k,m1)-r
    while True:
        m2=m1+dp
        f2=poissoncum(k,m2)-r
        if abs(f2)<eps:
            break
        if (abs(f2)>abs(f1)) or (f1*f2<0):
            dp=-dp/2
        m1=m2
        f1=f2
    return m2


###############################################################################
# Borne inférieure de confiance sur le paramètre m d'une distribution de Poisson
#
def poissonconfinf(k,conf):
    """poissonconfinf(k,conf): borne inférieure sur la moyenne trouvée avec un échantillon (k) et une confiance conf """
    if (not pmain.estnombre(k)) or (not pmain.estentier(k)) or (k<0):
        raise ValueError ("erreur poissonconfinf(k,conf): k doit être un nombre entier positif ou nul")
    if (not pmain.estnombre(conf)) or (conf<0) or (conf>1):
        raise ValueError ("erreur poissonconfinf(k,conf): conf doit être un nombre compris entre 0 et 1")
    dp=0.1
    eps=1.0e-8
    m1=0.5
    r=1-conf
    f1=1-poissoncum(k-1,m1)-r
    while True:
        m2=m1+dp
        f2=1-poissoncum(k-1,m2)-r
        if abs(f2)<eps:
            break
        if (abs(f2)>abs(f1)) or (f1*f2<0):
            dp=-dp/2
        m1=m2
        f1=f2
    return m2

###############################################################################
# Intervalle de confiance sur le paramètre m d'une distribution de Poisson
#
def poissonconf(k,conf):
    """poissonconf(k,n,conf): Intervalle de confiance sur le paramètre m d'une distribution de Poisson"""
    if (not pmain.estnombre(k)) or (not pmain.estentier(k)) or (k<0):
        raise ValueError ("erreur poissonconf(k,conf): k doit être un nombre entier positif ou nul")
    if (not pmain.estnombre(conf)) or (conf<0) or (conf>1):
        raise ValueError ("erreur poissonconf(k,conf): conf doit être un nombre compris entre 0 et 1")
    bi=poissonconfinf(k,conf+(1-conf)/2)
    bs=poissonconfsup(k,conf+(1-conf)/2)
    return [bi,bs]

###############################################################################
# Génération de valeurs tirées au hasard selon une distribution de Poisson de paramètre m
#
def hpoisson(m,nb=0):
    """hpoisson(m,nb=0): Génération de valeurs tirées au hasard selon une distribution de Poisson de paramètre m"""
    def _hpoisson(m):
        ph=random()
        k=0
        pc=poisson(k,m)
        while ph>pc:
            k+=1
            pc+=poisson(k,m)
        return k
    if (not pmain.estnombre(m)) or (m<0):
        raise ValueError ("erreur hpoisson(m,nb=0): m doit être un nombre positif ou nul")
    if (not pmain.estnombre(nb)) or (not pmain.estentier(nb)) or (nb<0):
        raise ValueError ("erreur hpoisson(m,nb=0): nb doit être un nombre entier positif ou nul")
    if nb==0:
        return _hpoisson(m)
    else:
        R=[]
        for j in xrange(0,nb):
            R.append(_hpoisson(m))
        return R

###############################################################################
# probabilité qu'une variable aléatoire distribuée selon une loi normale réduite soit inférieure à x
# au delà de x=7.56, renvoie 1.0
#
def pgaussred(x):
    """pgaussred(x): probabilité qu'une variable aléatoire distribuée selon une loi normale réduite soit inférieure à x"""
    if not pmain.estnombre(x):
        raise ValueError ("erreur pgaussred(x): x doit être un nombre")
    if x==0:
        return 0.5
    if x>7.56:
        return 1.0
    if x<-7.56:
        return 0.0
    u=abs(x)
    n=int(u*2000)
    du=u/n
    k=1/sqrt(2*pi)
    u1=0
    f1=k
    p=0.5
    for i in range(0,n):
        u2=u1+du
        f2=k*exp(-0.5*u2*u2)
        p=p+(f1+f2)*du*0.5
        u1=u2
        f1=f2
    if p>1.0:
        p=1.0
    if x>0:
        return p
    else:
        return 1.0-p

###############################################################################
# Loi normale (m, s)
#   prob qu'une variable aléatoire distribuée selon ln(m,s) soit inférieure à x
#
def pgauss(x,m,s):
    """pgauss(x): probabilité qu'une variable aléatoire distribuée selon une loi normale(m,s) soit inférieure à x"""
    if not pmain.estnombre(x):
        raise ValueError ("erreur pgauss(x,m,s): p doit être un nombre")
    if not pmain.estnombre(m):
        raise ValueError ("erreur pgauss(x,m,s): m doit être un nombre")
    if (not pmain.estnombre(s)) or (s<=0):
        raise ValueError ("erreur pgauss(x,m,s): s doit être un nombre positif non nul")
    return pgaussred((x-m)/s)

###############################################################################
# renvoie la variable x distribuée selon une LNR, et correspondant à une probabilité donnée p
#
def xgaussred(p):
    """xgaussred(p): renvoie la variable x distribuée selon une LNR, et correspondant à une probabilité donnée p"""
    # vérification des données
    if (not pmain.estnombre(p)) or (p<0) or (p>1):
        raise ValueError ("erreur xgaussred(p): p doit être un nombre >=0 et <=1")
    # calcul
    if p==0.5:
        return 0.0
    if p==0.0:
        return -7.56
    if p==1.0:
        return +7.56
    if p>0.5:
        pc=p
    else:
        pc=1-p
    du=1/2000
    eps=1.0E-15
    k=1/sqrt(2*pi)
    u1=0.0
    f1=k
    p1=0.5
    while True:
        u2=u1+du
        f2=k*exp(-0.5*u2*u2)
        s=(f1+f2)*0.5*du
        p2=p1+s
        if abs(p2-pc)<eps:
            break
        if ((p1<pc) and (p2<p1)) or  ((p1>pc) and (p2>p1)):
            # =on est en train de s'éloigner du point recherché
            du=-du/2
        u1=u2
        f1=f2
        p1=p2
    if p>0.5:
        return u2
    else:
        return -u2

###############################################################################
# renvoie la variable x distribuée selon une LN(m,s), et correspondant à une probabilité donnée p
#
def xgauss(p,m,s):
    """xgauss(p,m,s): renvoie la variable x distribuée selon une LN(m,s), et correspondant à une probabilité donnée p"""
    if (not pmain.estnombre(p)) or (p<0) or (p>1):
        raise ValueError ("erreur xgauss(p,m,s): p doit être un nombre >=0 et <=1")
    if not pmain.estnombre(m):
        raise ValueError ("erreur xgauss(p,m,s): m doit être un nombre")
    if not pmain.estnombre(s):
        raise ValueError ("erreur xgauss(p,m,s): s doit être un nombre")
    return m+s*xgaussred(p)

###############################################################################
# renvoie une valeur au hasard de la variable x distribuée selon une LN réduite
#
def hgaussred(n=0):
    """hgaussred(): renvoie une valeur au hasard de la variable x distribuée selon une LN réduite"""
    if (not pmain.estnombre(n)) or (not pmain.estentier(n)) or (n<0):
        raise ValueError ("erreur hgaussred(n=0): n doit être un nombre entier >=0")
    if n==0:
        return xgaussred(random())
    else:
        L=[]
        for i in range(0,n):
            L.append(xgaussred(random()))
        return L

###############################################################################
# renvoie une valeur au hasard de la variable x distribuée selon une LN(m,s)
# si nb=0 => renvoie une valeur unique, sinon, renvoie une liste de n valeur(s)
#
def hgauss(m,s,n=0):
    """hgauss(m,s): renvoie une valeur au hasard de la variable x distribuée selon une LN(m,s)"""
    if not pmain.estnombre(m):
        raise ValueError ("erreur hgauss(m,s,n=0): m doit être un nombre")
    if (not pmain.estnombre(s)) or (s<=0):
        raise ValueError ("erreur hgauss(m,s,n=0): s doit être un nombre positif non nul")
    if (not pmain.estnombre(n)) or (not pmain.estentier(n)) or (n<0):
        raise ValueError ("erreur hgauss(m,s,n=0): n doit être un nombre entier >=0")
    if n==0:
        return xgauss(random(),m,s)
    else:
        L=[]
        for i in range(0,n):
            L.append(xgauss(random(),m,s))
        return L

